import dash_leaflet as dl
import numpy as np
import pandas as pd
import time
import os
import dash_leaflet.express as dlx
from dash import dcc
import pandas as pd
import serial
import warnings
import serial.tools.list_ports
from dash_extensions.javascript import assign
from dash import Dash, html
from dash.dependencies import Input, Output

os.system("start \"\" http://127.0.0.1:8050/") # Open a web server on browser
chroma = "https://cdnjs.cloudflare.com/ajax/libs/chroma-js/2.1.0/chroma.min.js"  # js lib used for colors

# Connect to serial port
arduino_ports = [
	p.device
	for p in serial.tools.list_ports.comports()
	if 'Arduino' in p.description
]
if not arduino_ports:
	raise IOError("No Arduino found")
if len(arduino_ports) > 1:
	warnings.warn("Multiple Arduinos found - using the first")

# Create a colorbar
colorscale = ["white", "yellow", "red", "brown"] 
colorbar = dl.Colorbar(colorscale=colorscale, className="probability", width=30, height=200, min=0, max=1)

# Create the drone image
image_url = "/assets/drone.png"
dd = 0.0002
image_bounds = [[0,0], [0,0]]

# Pre process the data into geobuf.
cols = ['time', 'latitude', 'longitude', 'probability']
raggio_cerchio = 3
punto_corrente = [0, 0, 0, 0]
df=pd.DataFrame(columns=cols)
lista_coordinate = pd.DataFrame(columns=cols)
cc_df=pd.DataFrame(columns=np.hstack((cols, ['cumsum','n_meas'])))

dicts = df.to_dict('records')
for item in dicts:
    item["tooltip"] = f"time={item['time']}, latitude={item['latitude']}, longitude={item['longitude']}, probability={item['probability']}"
geojson = dlx.dicts_to_geojson(dicts, lat="latitude", lon="longitude")  # convert to geojson
geobuf = dlx.geojson_to_geobuf(geojson)  # convert to geobuf

# Geojson rendering logic, must be JavaScript as it is executed in clientside.
point_to_layer = assign("""function(feature, latlng, context){
    const {min, max, colorscale, circleOptions, colorProp} = context.props.hideout;
    const csc = chroma.scale(colorscale).domain([min, max]);  // chroma lib to construct colorscale
    circleOptions.fillColor = csc(feature.properties[colorProp]);  // set color based on color prop.
    var circle = L.circleMarker(latlng, circleOptions);
    return circle.bringToBack()  // sender a simple circle marker.
}""")
# Create geojson.
geojson = dl.GeoJSON(data=geobuf, id="geojson", format="geobuf",
                    zoomToBounds=True,  # when true, zooms to bounds when data changes
                    options=dict(pointToLayer=point_to_layer),  # how to draw points
                    superClusterOptions=dict(radius=50),   # adjust cluster size
                    hideout=dict(colorProp='probability', circleOptions=dict(fillOpacity=1, stroke=False, radius=8),
                                min=0, max=1, colorscale=colorscale))

# Create the app.
app = Dash(external_scripts=[chroma], prevent_initial_callbacks=True)
app.layout = html.Div([
                    html.Button(id="button", children="Start/Pause", className="button"),
                    dcc.Interval(
                        id='interval-component',
                        interval=1000, # in milliseconds
                        n_intervals=0,
                        disabled=True
                                ),    
                    dcc.Interval(
                        id='interval-component2',
                        interval=1000, # in milliseconds
                        n_intervals=0
                                ),      
                    html.Br(),
                    html.Hr(),
                    dl.Map([dl.ImageOverlay(url=image_url, bounds=image_bounds, id="drone", zIndex=1000), dl.TileLayer(), geojson, colorbar],
                                        bounds=image_bounds, id="map",
                                        style={'width': '100%', 'height': '80vh', 'margin': "auto", "display": "block"}
                          )
                    ], style={'text-align': 'center'})

def conversione_latlong_metri(cc_df, punto_corrente_ok):
    R = 6371e3  # radius of earth in meters
    lat1 = np.array(cc_df['latitude'])
    lon1 = np.array(cc_df['longitude'])
    lat2 = punto_corrente_ok[1]
    lon2 = punto_corrente_ok[2]
    phi1 = lat1 * np.pi / 180
    phi2 = lat2 * np.pi / 180
    delta_phi = (lat2 - lat1) * np.pi / 180
    delta_lambda = (lon2 - lon1) * np.pi / 180
    a = np.power(np.sin(delta_phi/2), 2) + np.cos(phi1) * np.cos(phi2) * np.power(np.sin(delta_lambda/2), 2)
    c = 2 * np.arctan2(np.sqrt(a), np.sqrt(1-a))
    return R * c    # in meters

@app.callback(
    Output('interval-component', 'disabled'),
    Input('button', "n_clicks")
)
def play_pause(n_clicks):
    if n_clicks % 2 == 0:
        cc_df[['latitude', 'longitude', 'probability']].to_csv('received_data.csv', index=False)
    return n_clicks % 2 == 0

@app.callback(
    Output('geojson', 'data'),
    [Input('interval-component', "n_intervals")]
)
def calcolo_punti(value):
    global df
    global punto_corrente_ok
    global lista_coordinate
    global cc_df

    punto_precedente = punto_corrente
    df_punto_corrente = pd.DataFrame([punto_corrente], columns=cols)
    if conversione_latlong_metri(df_punto_corrente, punto_precedente) <= raggio_cerchio:
        lista_coordinate = lista_coordinate.append(df_punto_corrente, ignore_index=True)
    else:
        lista_coordinate = pd.DataFrame(columns=cols)

    if (len(lista_coordinate) >= 5):
        punto_corrente_ok = [lista_coordinate.loc[0, 'time'], lista_coordinate.loc[0, 'latitude'], lista_coordinate.loc[0, 'longitude'], lista_coordinate['probability'].mean()]
        lista_coordinate = pd.DataFrame(columns=cols)
        distanza_metri = conversione_latlong_metri(cc_df, punto_corrente_ok)
        if np.all(distanza_metri >= 2 * raggio_cerchio):    # crea un nuovo punto sulla mappa
            cc_df = cc_df.append(pd.DataFrame([punto_corrente_ok + [0, 0]], columns=cc_df.columns), ignore_index=True)
        elif np.any(distanza_metri < raggio_cerchio):       # aggiorna il valore di probabilità del punto già esistente
            index_min = np.argmin(np.sqrt(np.power((cc_df.latitude - punto_corrente_ok[1]), 2) + np.power((cc_df.longitude - punto_corrente_ok[2]), 2)))
            cc_df.loc[index_min, 'cumsum'] = cc_df.loc[index_min, 'cumsum'] + punto_corrente_ok[3]
            cc_df.loc[index_min, 'n_meas'] = cc_df.loc[index_min, 'n_meas'] + 1
            cc_df.loc[index_min, 'probability'] = cc_df.loc[index_min, 'cumsum'] / cc_df.loc[index_min, 'n_meas']

    dicts = cc_df.to_dict('records')
    for item in dicts:
        item["tooltip"] = f"time={item['time']}, latitude={item['latitude']}, longitude={item['longitude']}, probability={item['probability']}"
    geojson = dlx.dicts_to_geojson(dicts, lat="latitude", lon="longitude")
    geobuf = dlx.geojson_to_geobuf(geojson)  # convert to geobuf

    return(geobuf)

@app.callback(
    Output('drone', 'bounds'),
    Output('map', 'bounds'),
    [Input('interval-component2', "n_intervals")]
)
def move_drone(n_intervals):
    global df
    global punto_corrente
    
    line = str(ser.readline())
    print(f"\n\n\n\n\n{line}")
    if line[:2] == "b\'":
        cc = line[2:][:-6]
        punto_corrente = cc.split(",")
        punto_corrente = [float(element) for element in punto_corrente]
        punto_corrente[3] = round(punto_corrente[3], 5)
        print(f"{punto_corrente}\n\n\n\n")

        df = df.append(pd.DataFrame([punto_corrente], columns=df.columns), ignore_index=True)
        image_bounds = [[df.iloc[len(df)-1].latitude - dd, df.iloc[len(df)-1].longitude - dd], [df.iloc[len(df)-1].latitude + dd, df.iloc[len(df)-1].longitude + dd]]

    return(image_bounds, image_bounds)

if __name__ == '__main__':
    app.run_server(port = 8050, dev_tools_ui=True,
              dev_tools_hot_reload=True, threaded=True)